import sys
from contextlib import contextmanager

import pytest
from dagster import file_relative_path
from dagster.core.host_representation import (
    ManagedGrpcPythonEnvRepositoryLocationOrigin,
    PipelineHandle,
)
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin


@contextmanager
def get_example_repository_location():
    loadable_target_origin = LoadableTargetOrigin(
        executable_path=sys.executable,
        python_file=file_relative_path(__file__, "repo.py"),
    )
    location_name = "example_repo_location"

    origin = ManagedGrpcPythonEnvRepositoryLocationOrigin(loadable_target_origin, location_name)

    with origin.create_test_location() as location:
        yield location


@pytest.fixture
def foo_example_repo():
    with get_example_repository_location() as location:
        yield location.get_repository("example_repo")


@pytest.fixture
def foo_pipeline_handle(foo_example_repo):  # pylint: disable=redefined-outer-name
    return PipelineHandle("foo_pipeline", foo_example_repo.handle)



import time

import pendulum
from dagster.daemon.controller import (
    DAEMON_HEARTBEAT_INTERVAL_SECONDS,
    DAEMON_HEARTBEAT_TOLERANCE_SECONDS,
    all_daemons_healthy,
)

from .utils import setup_instance, start_daemon


def test_heartbeat(
    tmpdir,
):

    dagster_home_path = tmpdir.strpath
    with setup_instance(dagster_home_path, "") as instance:

        assert all_daemons_healthy(instance) is False

        with start_daemon():
            time.sleep(5)
            assert all_daemons_healthy(instance) is True

        frozen_datetime = pendulum.now().add(
            seconds=DAEMON_HEARTBEAT_INTERVAL_SECONDS + DAEMON_HEARTBEAT_TOLERANCE_SECONDS + 5
        )
        with pendulum.test(frozen_datetime):
            assert all_daemons_healthy(instance) is False



import time

import pendulum
from dagster import DagsterInvariantViolationError
from dagster.core.test_utils import instance_for_test
from dagster.daemon.controller import (
    all_daemons_healthy,
    all_daemons_live,
    daemon_controller_from_instance,
    get_daemon_status,
)
from dagster.utils.error import SerializableErrorInfo


def test_healthy():

    with instance_for_test(
        overrides={
            "run_coordinator": {
                "module": "dagster.core.run_coordinator.queued_run_coordinator",
                "class": "QueuedRunCoordinator",
            },
        }
    ) as instance:
        init_time = pendulum.now("UTC")

        assert not all_daemons_healthy(instance, curr_time_seconds=init_time.float_timestamp)
        assert not all_daemons_live(instance, curr_time_seconds=init_time.float_timestamp)

        with daemon_controller_from_instance(instance) as controller:

            while True:
                now = pendulum.now("UTC")
                if all_daemons_healthy(
                    instance, curr_time_seconds=now.float_timestamp
                ) and all_daemons_live(instance, curr_time_seconds=now.float_timestamp):

                    controller.check_daemon_threads()
                    controller.check_daemon_heartbeats()

                    beyond_tolerated_time = now.float_timestamp + 100

                    assert not all_daemons_healthy(
                        instance, curr_time_seconds=beyond_tolerated_time
                    )
                    assert not all_daemons_live(instance, curr_time_seconds=beyond_tolerated_time)
                    break

                if (now - init_time).total_seconds() > 10:
                    raise Exception("timed out waiting for instance to become healthy")

                time.sleep(0.5)


def test_healthy_with_different_daemons():
    with instance_for_test() as instance:
        with daemon_controller_from_instance(instance):

            with instance_for_test(
                overrides={
                    "run_coordinator": {
                        "module": "dagster.core.run_coordinator.queued_run_coordinator",
                        "class": "QueuedRunCoordinator",
                    },
                }
            ) as other_instance:
                now = pendulum.now("UTC")
                assert not all_daemons_healthy(
                    other_instance, curr_time_seconds=now.float_timestamp
                )
                assert not all_daemons_live(other_instance, curr_time_seconds=now.float_timestamp)


def test_thread_die_daemon(monkeypatch):
    with instance_for_test(overrides={}) as instance:
        from dagster.daemon.daemon import SchedulerDaemon, SensorDaemon

        iteration_ran = {"ran": False}

        def run_iteration_error(_, _instance, _workspace):
            iteration_ran["ran"] = True
            raise KeyboardInterrupt
            yield  # pylint: disable=unreachable

        monkeypatch.setattr(SensorDaemon, "run_iteration", run_iteration_error)

        init_time = pendulum.now("UTC")
        with daemon_controller_from_instance(instance) as controller:
            while True:
                now = pendulum.now("UTC")

                status = get_daemon_status(
                    instance, SchedulerDaemon.daemon_type(), now.float_timestamp
                )

                if iteration_ran["ran"] and status.healthy:
                    try:
                        controller.check_daemon_threads()  # Should eventually throw since the sensor thread is interrupted
                    except Exception as e:  # pylint: disable=broad-except
                        assert (
                            "Stopping dagster-daemon process since the following threads are no longer running: ['SENSOR']"
                            in str(e)
                        )
                        break

                if (now - init_time).total_seconds() > 20:
                    raise Exception("timed out waiting for check_daemons to fail")

                time.sleep(0.5)


def test_error_daemon(monkeypatch):
    with instance_for_test() as instance:
        from dagster.daemon.daemon import SensorDaemon

        def run_iteration_error(_, _instance, _workspace):
            raise DagsterInvariantViolationError("foobar")
            yield  # pylint: disable=unreachable

        monkeypatch.setattr(SensorDaemon, "run_iteration", run_iteration_error)

        init_time = pendulum.now("UTC")
        with daemon_controller_from_instance(instance) as controller:
            while True:
                now = pendulum.now("UTC")

                if all_daemons_live(instance):
                    # Despite error, daemon should still be running
                    controller.check_daemon_threads()
                    controller.check_daemon_heartbeats()

                    status = get_daemon_status(
                        instance, SensorDaemon.daemon_type(), now.float_timestamp
                    )

                    assert status.healthy == False
                    assert len(status.last_heartbeat.errors) == 1
                    assert (
                        status.last_heartbeat.errors[0].message.strip()
                        == "dagster.core.errors.DagsterInvariantViolationError: foobar"
                    )
                    assert not all_daemons_healthy(instance, curr_time_seconds=now.float_timestamp)
                    assert all_daemons_live(instance, curr_time_seconds=now.float_timestamp)
                    break

                if (now - init_time).total_seconds() > 10:
                    raise Exception("timed out waiting for heartbeat error")

                time.sleep(0.5)


def test_multiple_error_daemon(monkeypatch):
    with instance_for_test() as instance:
        from dagster.daemon.daemon import SensorDaemon

        def run_iteration_error(_, _instance, _workspace):
            # ?message stack cls_name cause"
            yield SerializableErrorInfo("foobar", None, None, None)
            yield SerializableErrorInfo("bizbuz", None, None, None)

        monkeypatch.setattr(SensorDaemon, "run_iteration", run_iteration_error)

        init_time = pendulum.now("UTC")

        with daemon_controller_from_instance(instance) as controller:
            while True:

                now = pendulum.now("UTC")

                if all_daemons_live(instance):

                    # Despite error, daemon should still be running
                    controller.check_daemon_threads()
                    controller.check_daemon_heartbeats()

                    status = get_daemon_status(
                        instance, SensorDaemon.daemon_type(), now.float_timestamp
                    )

                    if status.healthy == False:
                        assert len(status.last_heartbeat.errors) == 2
                        assert status.last_heartbeat.errors[0].message.strip() == "foobar"
                        assert status.last_heartbeat.errors[1].message.strip() == "bizbuz"
                        break

                if (now - init_time).total_seconds() > 10:
                    raise Exception("timed out waiting for heartbeat error")

                time.sleep(0.5)


def test_warn_multiple_daemons(capsys):
    from dagster.daemon.daemon import SensorDaemon

    with instance_for_test() as instance:
        init_time = pendulum.now("UTC")

        with daemon_controller_from_instance(instance):
            while True:
                now = pendulum.now("UTC")

                if all_daemons_live(instance):
                    captured = capsys.readouterr()
                    assert "Taking over from another SENSOR daemon process" not in captured.out
                    break

                if (now - init_time).total_seconds() > 10:
                    raise Exception("timed out waiting for daemon status")

                time.sleep(0.5)

            capsys.readouterr()

        init_time = pendulum.now("UTC")

        status = get_daemon_status(instance, SensorDaemon.daemon_type(), now.float_timestamp)
        last_heartbeat_time = status.last_heartbeat.timestamp

        # No warning when a second controller starts up again
        with daemon_controller_from_instance(instance):
            while True:
                now = pendulum.now("UTC")

                status = get_daemon_status(
                    instance, SensorDaemon.daemon_type(), now.float_timestamp
                )

                if status.last_heartbeat and status.last_heartbeat.timestamp != last_heartbeat_time:
                    captured = capsys.readouterr()
                    assert "Taking over from another SENSOR daemon process" not in captured.out
                    break

                if (now - init_time).total_seconds() > 10:
                    raise Exception("timed out waiting for new daemon status")

                time.sleep(0.5)

            status = get_daemon_status(instance, SensorDaemon.daemon_type(), now.float_timestamp)
            last_heartbeat_time = status.last_heartbeat.timestamp

            # Starting up a controller while one is running produces the warning though
            with daemon_controller_from_instance(instance):
                # Wait for heartbeats while two controllers are running at once and there will
                # be a warning
                init_time = pendulum.now("UTC")

                while True:
                    now = pendulum.now("UTC")

                    captured = capsys.readouterr()
                    if "Taking over from another SENSOR daemon process" in captured.out:
                        break

                    if (now - init_time).total_seconds() > 60:
                        raise Exception("timed out waiting for heartbeats")

                    time.sleep(5)



import json

import pytest
from dagster import (
    AssetMaterialization,
    DagsterEventType,
    DagsterInvalidConfigError,
    InputDefinition,
    Int,
    Output,
    OutputDefinition,
    PipelineDefinition,
    String,
    dagster_type_materializer,
    execute_pipeline,
    lambda_solid,
    solid,
)
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution.plan.step import StepKind
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.types.dagster_type import create_any_type
from dagster.utils.test import get_temp_file_name, get_temp_file_names


def single_int_output_pipeline():
    @lambda_solid(output_def=OutputDefinition(Int))
    def return_one():
        return 1

    return PipelineDefinition(name="single_int_output_pipeline", solid_defs=[return_one])


def single_string_output_pipeline():
    @lambda_solid(output_def=OutputDefinition(String))
    def return_foo():
        return "foo"

    return PipelineDefinition(name="single_string_output_pipeline", solid_defs=[return_foo])


def multiple_output_pipeline():
    @solid(output_defs=[OutputDefinition(Int, "number"), OutputDefinition(String, "string")])
    def return_one_and_foo(_context):
        yield Output(1, "number")
        yield Output("foo", "string")

    return PipelineDefinition(name="multiple_output_pipeline", solid_defs=[return_one_and_foo])


def single_int_named_output_pipeline():
    @lambda_solid(output_def=OutputDefinition(Int, name="named"))
    def return_named_one():
        return Output(1, "named")

    return PipelineDefinition(
        name="single_int_named_output_pipeline", solid_defs=[return_named_one]
    )


def no_input_no_output_pipeline():
    @solid(output_defs=[])
    def take_nothing_return_nothing(_context):
        pass

    return PipelineDefinition(
        name="no_input_no_output_pipeline", solid_defs=[take_nothing_return_nothing]
    )


def one_input_no_output_pipeline():
    @solid(input_defs=[InputDefinition("dummy")], output_defs=[])
    def take_input_return_nothing(_context, **_kwargs):
        pass

    return PipelineDefinition(
        name="one_input_no_output_pipeline", solid_defs=[take_input_return_nothing]
    )


def test_basic_json_default_output_config_schema():
    env = EnvironmentConfig.build(
        single_int_output_pipeline(),
        {"solids": {"return_one": {"outputs": [{"result": {"json": {"path": "foo"}}}]}}},
    )

    assert env.solids["return_one"]
    assert env.solids["return_one"].outputs.type_materializer_specs == [
        {"result": {"json": {"path": "foo"}}}
    ]


def test_basic_json_named_output_config_schema():
    env = EnvironmentConfig.build(
        single_int_named_output_pipeline(),
        {"solids": {"return_named_one": {"outputs": [{"named": {"json": {"path": "foo"}}}]}}},
    )

    assert env.solids["return_named_one"]
    assert env.solids["return_named_one"].outputs.type_materializer_specs == [
        {"named": {"json": {"path": "foo"}}}
    ]


def test_basic_json_misnamed_output_config_schema():
    with pytest.raises(DagsterInvalidConfigError) as exc_context:
        EnvironmentConfig.build(
            single_int_named_output_pipeline(),
            {
                "solids": {
                    "return_named_one": {"outputs": [{"wrong_name": {"json": {"path": "foo"}}}]}
                }
            },
        )

    assert len(exc_context.value.errors) == 1
    assert 'Error 1: Received unexpected config entry "wrong_name"' in exc_context.value.message
    assert "at path root:solids:return_named_one:outputs[0]" in exc_context.value.message


def test_no_outputs_no_inputs_config_schema():
    assert EnvironmentConfig.build(no_input_no_output_pipeline())

    with pytest.raises(DagsterInvalidConfigError) as exc_context:
        EnvironmentConfig.build(no_input_no_output_pipeline(), {"solids": {"return_one": {}}})

    assert len(exc_context.value.errors) == 1
    assert (
        'Error 1: Received unexpected config entry "return_one" at path root:solids'
        in exc_context.value.message
    )


def test_no_outputs_one_input_config_schema():
    assert EnvironmentConfig.build(
        one_input_no_output_pipeline(),
        {"solids": {"take_input_return_nothing": {"inputs": {"dummy": {"value": "value"}}}}},
    )

    with pytest.raises(DagsterInvalidConfigError) as exc_context:
        EnvironmentConfig.build(
            one_input_no_output_pipeline(),
            {
                "solids": {
                    "take_input_return_nothing": {
                        "inputs": {"dummy": {"value": "value"}},
                        "outputs": {},
                    }
                }
            },
        )

    assert len(exc_context.value.errors) == 1
    exp_msg = 'Error 1: Received unexpected config entry "outputs" at path root:solids:take_input_return_nothing'
    assert exp_msg in exc_context.value.message


def test_basic_int_json_materialization():
    with get_temp_file_name() as filename:
        result = execute_pipeline(
            single_int_output_pipeline(),
            {"solids": {"return_one": {"outputs": [{"result": {"json": {"path": filename}}}]}}},
        )

        assert result.success

        with open(filename, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": 1}


def test_basic_materialization_event():
    with get_temp_file_name() as filename:
        result = execute_pipeline(
            single_int_output_pipeline(),
            {"solids": {"return_one": {"outputs": [{"result": {"json": {"path": filename}}}]}}},
        )

        assert result.success
        solid_result = result.result_for_solid("return_one")
        step_events = solid_result.step_events_by_kind[StepKind.COMPUTE]
        mat_event = list(
            filter(lambda de: de.event_type == DagsterEventType.ASSET_MATERIALIZATION, step_events)
        )[0]

        mat = mat_event.event_specific_data.materialization

        assert len(mat.metadata_entries) == 1
        assert mat.metadata_entries[0].path
        path = mat.metadata_entries[0].entry_data.path

        with open(path, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": 1}


def test_basic_string_json_materialization():
    pipeline = single_string_output_pipeline()

    with get_temp_file_name() as filename:
        result = execute_pipeline(
            pipeline,
            {"solids": {"return_foo": {"outputs": [{"result": {"json": {"path": filename}}}]}}},
        )

        assert result.success

        with open(filename, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": "foo"}


def test_basic_int_and_string_json_materialization():

    pipeline = multiple_output_pipeline()

    with get_temp_file_names(2) as file_tuple:
        filename_one, filename_two = file_tuple  # pylint: disable=E0632
        result = execute_pipeline(
            pipeline,
            {
                "solids": {
                    "return_one_and_foo": {
                        "outputs": [
                            {"string": {"json": {"path": filename_one}}},
                            {"number": {"json": {"path": filename_two}}},
                        ]
                    }
                }
            },
        )

        assert result.success

        with open(filename_one, "r") as ff_1:
            value = json.loads(ff_1.read())
            assert value == {"value": "foo"}

        with open(filename_two, "r") as ff_2:
            value = json.loads(ff_2.read())
            assert value == {"value": 1}


def read_file_contents(path):
    with open(path, "r") as ff:
        return ff.read()


def test_basic_int_and_string_json_multiple_materialization():

    pipeline = multiple_output_pipeline()

    with get_temp_file_names(4) as file_tuple:
        # False positive for unbalanced tuple unpacking
        # pylint: disable=E0632
        filename_one, filename_two, filename_three, filename_four = file_tuple
        result = execute_pipeline(
            pipeline,
            {
                "solids": {
                    "return_one_and_foo": {
                        "outputs": [
                            {"string": {"json": {"path": filename_one}}},
                            {"string": {"json": {"path": filename_two}}},
                            {"number": {"json": {"path": filename_three}}},
                            {"number": {"json": {"path": filename_four}}},
                        ]
                    }
                }
            },
        )

        assert result.success

        with open(filename_one, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": "foo"}

        with open(filename_two, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": "foo"}

        with open(filename_three, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": 1}

        with open(filename_four, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": 1}


def assert_step_before(steps, first_step, second_step):
    step_keys = [step.key for step in steps]
    assert step_keys.index(first_step) < step_keys.index(second_step)


def assert_plan_topological_level(steps, step_nums, step_keys):
    assert set(steps[step_num].key for step_num in step_nums) == set(step_keys)


def test_basic_int_json_multiple_materializations():
    pipeline = single_int_output_pipeline()

    with get_temp_file_names(2) as file_tuple:
        filename_one, filename_two = file_tuple  # pylint: disable=E0632
        result = execute_pipeline(
            pipeline,
            {
                "solids": {
                    "return_one": {
                        "outputs": [
                            {"result": {"json": {"path": filename_one}}},
                            {"result": {"json": {"path": filename_two}}},
                        ]
                    }
                }
            },
        )

        assert result.success

        with open(filename_one, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": 1}

        with open(filename_two, "r") as ff:
            value = json.loads(ff.read())
            assert value == {"value": 1}


@dagster_type_materializer(Int)
def yield_two_materializations(*_args, **_kwargs):
    yield AssetMaterialization("first")
    yield AssetMaterialization("second")


def test_basic_yield_multiple_materializations():
    SomeDagsterType = create_any_type(name="SomeType", materializer=yield_two_materializations)

    @lambda_solid(output_def=OutputDefinition(SomeDagsterType))
    def return_one():
        return 1

    pipeline_def = PipelineDefinition(name="single_int_output_pipeline", solid_defs=[return_one])
    result = execute_pipeline(
        pipeline_def, run_config={"solids": {"return_one": {"outputs": [{"result": 2}]}}}
    )
    assert result.success

    event_types = [event.event_type_value for event in result.event_list]
    assert 2 == (
        sum(
            [
                True
                for event_type in event_types
                if event_type == DagsterEventType.ASSET_MATERIALIZATION.value
            ]
        )
    )


@dagster_type_materializer(Int)
def return_int(*_args, **_kwargs):
    return 1


def test_basic_bad_output_materialization():
    SomeDagsterType = create_any_type(name="SomeType", materializer=return_int)

    @lambda_solid(output_def=OutputDefinition(SomeDagsterType))
    def return_one():
        return 1

    pipeline_def = PipelineDefinition(name="single_int_output_pipeline", solid_defs=[return_one])

    with pytest.raises(
        DagsterInvariantViolationError, match="You must return an AssetMaterialization"
    ):
        execute_pipeline(
            pipeline_def, run_config={"solids": {"return_one": {"outputs": [{"result": 2}]}}}
        )



from collections import namedtuple
from typing import Any, Optional

from dagster import check
from dagster.core.definitions.logger import LoggerDefinition
from dagster.core.definitions.pipeline import PipelineDefinition


class InitLoggerContext(
    namedtuple("InitLoggerContext", "logger_config pipeline_def logger_def run_id")
):
    """Logger-specific initialization context.

    An instance of this class is made available as the first argument to the ``logger_fn`` decorated
    by :py:func:`@logger <logger>` or set on a :py:class:`LoggerDefinition`.

    Users should not instantiate this class.

    Attributes:
        logger_config (Any): The configuration data provided by the environment config. The
            schema for this data is defined by ``config_schema`` on the :py:class:`LoggerDefinition`
        pipeline_def (Optional[PipelineDefinition]): The pipeline definition currently being executed.
        logger_def (LoggerDefinition): The logger definition for the logger being constructed.
        run_id (str): The ID for this run of the pipeline.
    """

    def __new__(
        cls,
        logger_config: Any,
        logger_def: LoggerDefinition,
        pipeline_def: Optional[PipelineDefinition] = None,
        run_id: Optional[str] = None,
    ):
        return super(InitLoggerContext, cls).__new__(
            cls,
            logger_config,
            check.opt_inst_param(pipeline_def, "pipeline_def", PipelineDefinition),
            check.inst_param(logger_def, "logger_def", LoggerDefinition),
            check.opt_str_param(run_id, "run_id"),
        )
